C
C  This program reads phase-space files (PSFs) and generates histograms
C  of the energy distribution and the spatial distribution of a selected
C  kind of particles from these PDFs. The energy distribution is set by
C  considering only particles with positions within the selected spatial
C  box.
C
C  To run the program, place the executable binary 'viewpsf.exe' and the
C  input file 'view.in' in the same working directory. Edit the input
C  file 'view.in' to declare the relative path and name of each PSF, and
C  to define the details of the histograms. To run the program, issue
C  the command 'viewpsf.exe < view.in'; five output files with the
C  calculated histograms are written on the working directory. If you
C  have gnuplot installed on your computer, the scripts '*.gnu' display
C  the histograms on the screen.
C
C                                        Francesc Salvat. August, 2014.
C


C  >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      MODULE TRACK_mod   ! Particle TRACK variables.
C  ----  Energy, position, direction, and weight.
      DOUBLE PRECISION :: E,X,Y,Z,U,V,W,WGHT
C  ----  Particle type, current body (geometry parameter), and material.
      INTEGER*4 :: KPAR,IBODY,MAT
C  ----  Particle history flags.
      INTEGER*4, DIMENSION (5) :: ILB
C  ****  Photon polarisation
C  ----  Polarised photons if IPOL=1, otherwise unpolarised photons.
      INTEGER*4 :: IPOL=0
C  ----  Stokes parameters.
      DOUBLE PRECISION :: SP1,SP2,SP3
C  ****  The particle age (time elapsed since the start of the shower)
C        is recorded when LAGE=.TRUE.
      LOGICAL :: LAGE =.FALSE.
      DOUBLE PRECISION :: PAGE=0.0D0
      END MODULE TRACK_mod
C  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


C  *********************************************************************
C                       MAIN PROGRAM
C  *********************************************************************
C
      USE TRACK_mod
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      CHARACTER*45 PSFN(50)
      PARAMETER (FSAFE=1.000000001D0)  ! Safety factor.
C  ----  Energy distribution.
      PARAMETER (NEM=1000)
      COMMON/CPDFE/PE(NEM),PE2(NEM),PEP(NEM),LPE(NEM),NE
C  ----  Angular distributions.
      PARAMETER (PI=3.1415926535897932D0, TWOPI=PI+PI,
     1  DE2RA=PI/180.0D0, RA2DE=180.0D0/PI)
      PARAMETER (NBTHM=180,NBPHM=360)  ! Max. nos of angular bins.
      COMMON/CENANG/PDA(NBTHM,NBPHM),PDA2(NBTHM,NBPHM),
     1   PDAP(NBTHM,NBPHM),LPDA(NBTHM,NBPHM),NTH,NPH
C
C
      READ(5,*) NPSF
      DO I=1,NPSF
        READ(5,'(A45)') PSFN(I)
      ENDDO
C
      READ(5,*) XL,XU,NBX
      READ(5,*) YL,YU,NBY
      READ(5,*) ZL,ZU,NBZ
C
      READ(5,*) EL,EU,NE
      NE=MAX(10,MIN(ABS(NE),NEM))
C
      READ(5,*) KPARS
C
C  ****  Energy distribution. Initialization.
C
      BE=FSAFE*(EU-EL)/DBLE(NE)
      RBE=1.0D0/BE
      DO IE=1,NEM
        PE(IE)=0.0D0
        PE2(IE)=0.0D0
        PEP(IE)=0.0D0
        LPE(IE)=0
      ENDDO
C
C  ****  Angular distribution. Initialization.
C
      NTH=90
      NPH=60
      BSTH=FSAFE*180.0D0/DBLE(NTH)
      RBSTH=1.0D0/BSTH
      BSPH=FSAFE*360.0D0/DBLE(NPH)
      RBSPH=1.0D0/BSPH
      DO J=1,NBTHM
        DO K=1,NBPHM
          PDA(J,K)=0.0D0
          PDA2(J,K)=0.0D0
          PDAP(J,K)=0.0D0
          LPDA(J,K)=0
        ENDDO
      ENDDO
C
C  ****  Map of initial positions of particles in the PSF
C
      CALL Q3D0(XL,XU,YL,YU,ZL,ZU,NBX,NBY,NBZ)  ! Initialization.
C
C
      IUNIT=21
      IPSF=0
      SHN=0.0D0
      N=0
C
  100 CONTINUE
      IPSF=IPSF+1
      IF(IPSF.LE.NPSF) THEN
        OPEN(IUNIT,FILE=PSFN(IPSF),IOSTAT=KODE,STATUS='OLD')
        IF(KODE.NE.0) THEN
          WRITE(6,'(''File '',A25,'' could not be opened.'')')
     1      PSFN(IPSF)
          IF(IPSF.LT.NPSF) GO TO 100
        ELSE
          WRITE(6,1001) PSFN(IPSF)
 1001     FORMAT(/3X,'+ The phase-space file ',A25,' is opened.')
        ENDIF
      ELSE
        GO TO 300
      ENDIF
      KODEPS=0
C
C  ************  Load a new particle from the PSF.
C
  200 CONTINUE
      CALL RDPSF(IUNIT,NSHI,ISEC,KODEPS)
      IF(KODEPS.EQ.-1) THEN  ! End of the PSF.
        CLOSE(IUNIT)
        WRITE(6,1002) SHN
 1002   FORMAT(3X,'Number of simulated showers =',1P,E14.7)
        GO TO 100
      ENDIF
C
      SHN=SHN+DBLE(NSHI)
      N=N+NSHI
      IF(N.GT.2000000000) N=N-2000000000
      IF(KPAR.NE.KPARS) GO TO 200
C
C  ************  Scoring.
C
      CALL SQ3D(WGHT,X,Y,Z,N)
C
      IE=1.0D0+(E-EL)*RBE
      IF(IE.GT.0.AND.IE.LE.NE) THEN
        IF(N.NE.LPE(IE)) THEN
          PE(IE)=PE(IE)+PEP(IE)
          PE2(IE)=PE2(IE)+PEP(IE)**2
          PEP(IE)=WGHT
          LPE(IE)=N
        ELSE
          PEP(IE)=PEP(IE)+WGHT
        ENDIF
      ENDIF
C
      THETA=ACOS(W)
      KTH=1.0D0+THETA*RA2DE*RBSTH
      IF(ABS(U).GT.1.0D-16) THEN  ! Azimuthal bin number corrected.
         PHI=ATAN2(V,U)
      ELSE IF(ABS(V).GT.1.0D-16) THEN
         PHI=ATAN2(V,U)
      ELSE
         PHI=0.0D0
      ENDIF
      IF(PHI.LT.0.0D0) PHI=TWOPI+PHI
      KPH=1.0D0+PHI*RA2DE*RBSPH
      IF(N.NE.LPDA(KTH,KPH)) THEN
        PDA(KTH,KPH)=PDA(KTH,KPH)+PDAP(KTH,KPH)
        PDA2(KTH,KPH)=PDA2(KTH,KPH)+PDAP(KTH,KPH)**2
        PDAP(KTH,KPH)=WGHT
        LPDA(KTH,KPH)=N
      ELSE
        PDAP(KTH,KPH)=PDAP(KTH,KPH)+WGHT
      ENDIF
C
      GO TO 200
C
C  ************  Write output files.
C
  300 CONTINUE
C
C  ****  Transfer partial counters to global counters.
C
      DO IE=1,NE
        PE(IE)=PE(IE)+PEP(IE)
        PE2(IE)=PE2(IE)+PEP(IE)**2
      ENDDO
C
      DO KTH=1,NTH
        DO KPH=1,NPH
          PDA(KTH,KPH)=PDA(KTH,KPH)+PDAP(KTH,KPH)
          PDA2(KTH,KPH)=PDA2(KTH,KPH)+PDAP(KTH,KPH)**2
        ENDDO
      ENDDO
C
      DF=1.0D0/SHN
C
      CALL Q3DW(SHN)
C
      OPEN(9,FILE='psf-energy.dat')
      WRITE(9,2001) KPARS
 2001 FORMAT(
     1   1X,'#  Results from VIEW-PSF',
     1  /1X,'#  Energy distribution of particles KPAR=',I2,
     1  /1X,'#  1st column: energy (eV).',
     1  /1X,'#  2nd column: PDF(E) (1/eV).',
     1  /1X,'#  3rd column: statistical uncertainty, 3 sigma /1/eV).')
      DO IE=1,NE
        E=EL+(IE-0.5D0)*BE
        YAV=PE(IE)
        YAV2=PE2(IE)
        YERR=3.0D0*SQRT(ABS(YAV2-YAV**2*DF))
        YAV=YAV*DF*RBE
        YERR=YERR*DF*RBE
        WRITE(9,'(1X,1P,3E14.6)') E,MAX(YAV,1.0D-35),YERR
      ENDDO
      CLOSE(9)
C
      OPEN(9,FILE='psf-angle.dat')
      WRITE(9,9210) KPARS
 9210 FORMAT(
     11X,'#  Results from VIEW-PSF.',
     1 /1X,'#  Angular distribution of particles KPAR=',I2,
     1 /1X,'#  1st and 2nd columns: THETA and PHI (deg).',
     1 /1X,'#  3rd and 4th columns: PDF and STU.',
     1 /1X,'#  PDFs and STUs in units of 1/(sr*primary_particle).')
      DO KTH=1,NTH
        XLOW=(KTH-1)*BSTH
        XUPP=KTH*BSTH
        XX=0.5D0*(XUPP+XLOW)
        BINS=XUPP-XLOW
        DSANG=(COS(XLOW*DE2RA)-COS(XUPP*DE2RA))*(BSPH*DE2RA)
        DO L=1,NPH
          YY=(L-0.5D0)*BSPH
          YERR1=3.0D0*SQRT(ABS(PDA2(KTH,L)-PDA(KTH,L)**2*DF))
          YAV1=PDA(KTH,L)*DF/DSANG
          YERR1=YERR1*DF/DSANG
          WRITE(9,'(1P,2E14.6,3(E14.6,E10.2))')
     1      XX,YY,MAX(YAV1,1.0D-35),YERR1
        ENDDO
        WRITE(9,*) '   '
      ENDDO
      CLOSE(9)
C
      WRITE(6,*) '  ***  Calculation completed successfully.'
      END
C  *********************************************************************
C                       SUBROUTINE RDPSF
C  *********************************************************************
      SUBROUTINE RDPSF(IUNIT,NSHI,ISEC,KODEPS)
C
C  This subroutine reads the phase-space file (psf). Blank lines and
C  lines starting with the pound sign (#) are considered as comments and
C  are ignored.
C
C  The reading of a new psf is initiated by calling subroutine RDPSF
C  with KODEPS=0. If the psf contains some particles, the oputput value
C  of KODEPS is set equal to 1. In subsequent calls to subroutine RDPSF,
C  the output value KODEPS=1 indicates that a valid record has been read
C  and that the state variables of a particle have been loaded in common
C  TRACK. The output value of ISEC is 1 if the psf contains more
C  particles from the current shower, and 0 otherwise. That is, when
C  ISEC=0 the next particle in the psf belongs to a new shower. The
C  output value KODEPS=-1 indicates that all particle records in the psf
C  have been read (or that the psf was empty). Note that when KODEPS=-1,
C  the contents of module TRACK_mod remains unchanged.
C
C  To reduce disk-access time, particle records are read in bunches and
C  stored in a buffer. The buffer size (number of particles in a bunch)
C  is defined by the parameter NRBUFF.
C
C  To verify that the input psfs are read correctly, uncomment the lines
C  with 'CALL WRPSF' (which write the read particle records in an output
C  file, UNIT=13). You should also add a sentence 'CALL WRPSF(13,0,1)'
C  at the end of the main program, which forces the writing of particle
C  records that may remain in the buffer and closes the output file.
C  This file should contain all the records read from the psf files.
C
      USE TRACK_mod
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
C
      PARAMETER (NRBUFF=1000)  ! Buffer size.
      CHARACTER PSFR(NRBUFF)*136
      SAVE PSFR
C
      SAVE E0,X0,Y0,Z0,U0,V0,W0,WGHT0,KPAR0,ILB10,ILB20,ILB30,ILB40,
     1  NSHI0,ILAST,IRD,IEOF
C
C  ****  Initialization (first call after opening the psf).
C
      IF(KODEPS.EQ.0) THEN
        NSHI=0
        ISEC=0
        IEOF=0
        DO I=1,NRBUFF
         PSFR(I)='#'  ! Identifies empty records.
        ENDDO
        READ(IUNIT,'(A)',END=10,ERR=10) (PSFR(I),I=1,NRBUFF)
        GO TO 11
   10   CONTINUE
        IEOF=1
   11   CONTINUE
C
        IRD=0
   20   CONTINUE
        IF(IRD.GE.NRBUFF) THEN
          IF(IEOF.EQ.1) THEN
            KODEPS=-1  ! End of file.
            RETURN
          ENDIF
          DO I=1,NRBUFF
           PSFR(I)='#'  ! Identifies empty records.
          ENDDO
          READ(IUNIT,'(A)',END=21,ERR=21) (PSFR(I),I=1,NRBUFF)
          GO TO 22
   21     CONTINUE
          IEOF=1
   22     CONTINUE
          IRD=0
        ENDIF
        IRD=IRD+1
        READ(PSFR(IRD),*,ERR=20,END=20) KPAR0,E0,X0,Y0,Z0,U0,V0,W0,
     1    WGHT0,ILB10,ILB20,ILB30,ILB40,NSHI0
        ILAST=0
        KODEPS=1
        RETURN
      ENDIF
C
C  ****  The last particle in the psf has already been loaded.
C
      IF(ILAST.EQ.1) THEN
        NSHI=0
        ISEC=0
        KODEPS=-1  ! End of file.
C       CALL WRPSF(13,NSHI,-1)  ! Uncomment to check psf reading.
        RETURN
      ENDIF
C
C  ****  A read particle is loaded.
C
      E=E0
      X=X0
      Y=Y0
      Z=Z0
      U=U0
      V=V0
      W=W0
      WGHT=WGHT0
      KPAR=KPAR0
      ILB(1)=ILB10
      ILB(2)=ILB20
      ILB(3)=ILB30
      ILB(4)=ILB40
      NSHI=NSHI0
C     CALL WRPSF(13,NSHI,0)  ! Uncomment to check psf reading.
      KODEPS=1
C
C  ****  Read a new particle and keep its state variables in memory.
C        This is needed to identify possible secondary particles that
C        belong to the same shower.
C
   30 CONTINUE
      IF(IRD.GE.NRBUFF) THEN
        IF(IEOF.EQ.1) THEN
          ILAST=1
          ISEC=0
          RETURN
        ENDIF
        DO I=1,NRBUFF
         PSFR(I)='#'  ! Identifies empty records.
        ENDDO
        READ(IUNIT,'(A)',END=31,ERR=31) (PSFR(I),I=1,NRBUFF)
        GO TO 32
   31   CONTINUE
        IEOF=1
   32   CONTINUE
        IRD=0
      ENDIF
      IRD=IRD+1
      READ(PSFR(IRD),*,ERR=30,END=30) KPAR0,E0,X0,Y0,Z0,U0,V0,W0,
     1    WGHT0,ILB10,ILB20,ILB30,ILB40,NSHI0
C
      IF(NSHI0.EQ.0) THEN
        ISEC=1
      ELSE
        ISEC=0
      ENDIF
      RETURN
      END
C  *********************************************************************
C                       SUBROUTINE WRPSF
C  *********************************************************************
      SUBROUTINE WRPSF(IPSFO,NSHI,ICLOSE)
C
C  This subroutine writes particle records in the output phase-space
C  file (psf). To reduce disk-access time, particle records are
C  collected in a buffer and moved to the psf when the buffer is full.
C  The buffer size (number of records) is set by the parameter NRBUFF.
C
C  Particles that remain in the buffer at the end of a simulation run
C  must be transferred to the psf by calling subroutine WRPSF with
C  ICLOSE=1.
C
C  Input parameters:
C     IPSFO ..... output unit.
C     NSHI ...... incremental shower number.
C     ICLOSE .... closing flag, acts only when its value is not 0.
C                 When ICLOSE .GT. 0, particles remaining in the buffer
C                 are transferred to the psf, and the output unit IPSFO
C                 is closed.
C                 If ICLOSE .LT. 0, particles in the buffer are moved to
C                 the psf, but the output unit remains open.
C
      USE TRACK_mod
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      CHARACTER*10 LC10A,LC10B
C
      PARAMETER (NRBUFF=1000)  ! Buffer size.
      CHARACTER PSFR(NRBUFF)*136
      SAVE PSFR
C
      DATA NREC/0/
      SAVE NREC
C
      IF(ICLOSE.NE.0) THEN
        IF(NREC.GT.0) THEN
          DO I=1,NREC
            NCHAR=INDEX(PSFR(I),'*')-1
            WRITE(IPSFO,'(A)') PSFR(I)(1:NCHAR)
          ENDDO
        ENDIF
        IF(ICLOSE.GT.0) CLOSE(IPSFO)
        NREC=0
        RETURN
      ENDIF
C
      NREC=NREC+1
      CALL N2CH10(ILB(4),LC10A,NDIGA)
      CALL N2CH10(NSHI,LC10B,NDIGB)
      WRITE(PSFR(NREC),'(I2,1P,8E13.5,I3,I2,I2,1X,A,1X,A,A)')
     1  KPAR,E,X,Y,Z,U,V,W,WGHT,ILB(1),ILB(2),ILB(3),
     1  LC10A(1:NDIGA),LC10B(1:NDIGB),'*'
C
      IF(NREC.EQ.NRBUFF) THEN
        WRITE(IPSFO,'(A)') (PSFR(I)(1:INDEX(PSFR(I),'*')-1),I=1,NREC)
        NREC=0
      ENDIF
C
      RETURN
      END
C  *********************************************************************
C                       SUBROUTINE N2CH10
C  *********************************************************************
      SUBROUTINE N2CH10(N,L,NDIG)
C
C  This subroutine writes an integer number N in a 10-character string
C  L. The number is written at the left, followed by unused blanks. NDIG
C  is the number of decimal digits of N.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      CHARACTER*10 L,LT
C
      WRITE(LT,'(I10)') N
      DO I=1,10
        IF(LT(I:I).NE.' ') THEN
          IT=I-1
          GO TO 1
        ENDIF
      ENDDO
      IT=9
    1 CONTINUE
      L=LT(IT+1:10)
      NDIG=10-IT
      RETURN
      END


C  *********************************************************************
C                       SUBROUTINE SQ3D
C  *********************************************************************
      SUBROUTINE SQ3D(Q,XD,YD,ZD,N)
C
C  Tallies the distribution of a continuous quantity Q in the volume
C  of a box, writes and loads dump files, accumulates dump files from
C  different runs, and writes results.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      PARAMETER (FSAFE=1.000000001D0)  ! Safety factor.
C  ----  Dose distribution (up to NDXM*NDYM*NDZM volume bins).
      PARAMETER (NDXM=101,NDYM=101,NDZM=101)
      COMMON/CMAP31/DQ(NDXM,NDYM,NDZM),DQ2(NDXM,NDYM,NDZM),
     1  DQP(NDXM,NDYM,NDZM),LQ(NDXM,NDYM,NDZM)
      COMMON/CMAP32/DXL(3),DXU(3),BQ(3),RBQ(3),NQB(3)
C
C  ************  3D distribution.
C
      IF((XD.GT.DXL(1).AND.XD.LT.DXU(1)).AND.
     1   (YD.GT.DXL(2).AND.YD.LT.DXU(2)).AND.
     1   (ZD.GT.DXL(3).AND.ZD.LT.DXU(3))) THEN
        I1=1.0D0+(XD-DXL(1))*RBQ(1)
        I2=1.0D0+(YD-DXL(2))*RBQ(2)
        I3=1.0D0+(ZD-DXL(3))*RBQ(3)
        IF(N.NE.LQ(I1,I2,I3)) THEN
          DQ(I1,I2,I3)=DQ(I1,I2,I3)+DQP(I1,I2,I3)
          DQ2(I1,I2,I3)=DQ2(I1,I2,I3)+DQP(I1,I2,I3)**2
          DQP(I1,I2,I3)=Q
          LQ(I1,I2,I3)=N
        ELSE
          DQP(I1,I2,I3)=DQP(I1,I2,I3)+Q
        ENDIF
      ENDIF
      RETURN
C
C  ************  Initialize the Q tallies.
C
      ENTRY Q3D0(XL,XU,YL,YU,ZL,ZU,NBX,NBY,NBZ)  !<<<<<<<<<<<<<<<<<<<<<<
C
        IF(NBX.LT.0.OR.NBX.GT.NDXM) THEN
          WRITE(26,'(''NBX ='',I6)') NBX
          WRITE(26,'(''NBX must be .GT.0. and .LT.'',I4)') NDXM
          WRITE(26,*) 'Increase the value of the parameter NDXM.'
          STOP 'SMAP3: NBX must be .GT.0. and .LE.NDXM'
        ENDIF
        IF(NBY.LT.0.OR.NBY.GT.NDYM) THEN
          WRITE(26,'(''NBY ='',I6)') NBY
          WRITE(26,'(''NBY must be .GT.0. and .LT.'',I4)') NDYM
          WRITE(26,*) 'Increase the value of the parameter NDYM.'
          STOP 'SMAP3: NBY must be .GT.0. and .LE.NDYM'
        ENDIF
        IF(NBZ.LT.0.OR.NBZ.GT.NDZM) THEN
          WRITE(26,'(''NBZ ='',I6)') NBZ
          WRITE(26,'(''NBZ must be .GT.0. and .LT.'',I4)') NDZM
          WRITE(26,*) 'Increase the value of the parameter NDZM.'
          STOP 'SMAP3: NBZ must be .GT.0. and .LE.NDZM'
        ENDIF
C
        NQB(1)=2*(NBX/2)+1
        NQB(2)=2*(NBY/2)+1
        NQB(3)=2*(NBZ/2)+1
        DXL(1)=XL
        DXU(1)=XU
        DXL(2)=YL
        DXU(2)=YU
        DXL(3)=ZL
        DXU(3)=ZU
C
        DO I=1,3
          BQ(I)=FSAFE*(DXU(I)-DXL(I))/DBLE(NQB(I))
          RBQ(I)=1.0D0/BQ(I)
        ENDDO
C
        DO K=1,NDZM
          DO J=1,NDYM
            DO I=1,NDXM
              DQ(I,J,K)=0.0D0
              DQ2(I,J,K)=0.0D0
              DQP(I,J,K)=0.0D0
              LQ(I,J,K)=0
            ENDDO
          ENDDO
        ENDDO
      RETURN
C
C  ************  DUMP. Write the histograms.
C
      ENTRY Q3DD(IWR)  !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
C
C  *****  Transfer partial counters to global counters.
C
        DO I3=1,NQB(3)
        DO I2=1,NQB(2)
        DO I1=1,NQB(1)
          DQ(I1,I2,I3)=DQ(I1,I2,I3)+DQP(I1,I2,I3)
          DQ2(I1,I2,I3)=DQ2(I1,I2,I3)+DQP(I1,I2,I3)**2
          DQP(I1,I2,I3)=0.0D0
          LQ(I1,I2,I3)=0
        ENDDO
        ENDDO
        ENDDO
C
        WRITE(IWR,*) (NQB(I),I=1,3)
        WRITE(IWR,*) (DXL(I),I=1,3)
        WRITE(IWR,*) (BQ(I),I=1,3)
        WRITE(IWR,*)
     1    (((DQ(I1,I2,I3),I3=1,NQB(3)),I2=1,NQB(2)),I1=1,NQB(1))
        WRITE(IWR,*)
     1    (((DQ2(I1,I2,I3),I3=1,NQB(3)),I2=1,NQB(2)),I1=1,NQB(1))
      RETURN
C
C  ************  RESUME. Read histograms from a previous run.
C
      ENTRY Q3DR(IRD)  !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
C
        READ(IRD,*) (NQB(I),I=1,3)
        READ(IRD,*) (DXL(I),I=1,3)
        READ(IRD,*) (BQ(I),I=1,3)
        READ(IRD,*)
     1    (((DQ(I1,I2,I3),I3=1,NQB(3)),I2=1,NQB(2)),I1=1,NQB(1))
        READ(IRD,*)
     1    (((DQ(I1,I2,I3),I3=1,NQB(3)),I2=1,NQB(2)),I1=1,NQB(1))
      RETURN
C
C  ************  ADD results from previous runs.
C  Use Q3DR to load histograms from one of the runs and then call
C  Q3DA to accumulate results from other runs.
C
      ENTRY Q3DA(IRD)  !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
C
        READ(IRD,*) (NQB(I),I=1,3)
        READ(IRD,*) (DXL(I),I=1,3)
        READ(IRD,*) (BQ(I),I=1,3)
        READ(IRD,*)
     1    (((DQP(I1,I2,I3),I3=1,NQB(3)),I2=1,NQB(2)),I1=1,NQB(1))
        DO I1=1,NQB(1)
          DO I2=1,NQB(2)
            DO I3=1,NQB(3)
              DQ(I1,I2,I3)=DQ(I1,I2,I3)+DQP(I1,I2,I3)
            ENDDO
          ENDDO
        ENDDO
        READ(IRD,*)
     1    (((DQP(I1,I2,I3),I3=1,NQB(3)),I2=1,NQB(2)),I1=1,NQB(1))
        DO I1=1,NQB(1)
          DO I2=1,NQB(2)
            DO I3=1,NQB(3)
              DQ(I1,I2,I3)=DQ(I1,I2,I3)+DQP(I1,I2,I3)
              DQP(I1,I2,I3)=0.0D0
            ENDDO
          ENDDO
        ENDDO
      RETURN
C
C  ************  Write simulation results in output files.
C
      ENTRY Q3DW(SHN)  !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        DF=1.0D0/SHN
C
C  *****  Transfer partial counters to global counters.
C
        DO I3=1,NQB(3)
        DO I2=1,NQB(2)
        DO I1=1,NQB(1)
          DQ(I1,I2,I3)=DQ(I1,I2,I3)+DQP(I1,I2,I3)
          DQ2(I1,I2,I3)=DQ2(I1,I2,I3)+DQP(I1,I2,I3)**2
          DQP(I1,I2,I3)=0.0D0
          LQ(I1,I2,I3)=0
        ENDDO
        ENDDO
        ENDDO
C
        VOXEL=BQ(1)*BQ(2)*BQ(3)
C
        OPEN(9,FILE='psf-3d-map.dat')
        WRITE(9,9420)
 9420   FORMAT(1X,'#  Results from PENMAIN. Q distribution (3D).')
        WRITE(9,9421) DXL(1),DXU(1)
 9421   FORMAT(1X,'#  Dose-map box:  XL = ',1P,E13.6,
     1    ' cm,  XU = ',E13.6,' cm')
        WRITE(9,9422) DXL(2),DXU(2)
 9422   FORMAT(1X,'#',17X,'YL = ',1P,E13.6,' cm,  YU = ',E13.6,' cm')
        WRITE(9,9423) DXL(3),DXU(3)
 9423   FORMAT(1X,'#',17X,'ZL = ',1P,E13.6,' cm,  ZU = ',E13.6,' cm')
        WRITE(9,9424) NQB(1),NQB(2),NQB(3)
 9424   FORMAT(1X,'#  Numbers of bins:     NBX =',I4,', NBY =',I4,
     1        ', NBZ =',I4,/1X,'#')
        WRITE(9,9425)
 9425   FORMAT(1X,'#  columns 1 to 3: coordinates X,Y,Z of the bin',
     1    ' centres',/1X,'#  4th column: Q (1/cm^3).',
     1    /1X,'#  5th column: statistical uncertainty (3 sigma).',
     1    /1X,'#  columns 6 to 8: bin indices IX,IY,IZ.')
        DO I3=1,NQB(3)
          ZZ=DXL(3)+(I3-0.5D0)*BQ(3)
          DO I1=1,NQB(1)
            XX=DXL(1)+(I1-0.5D0)*BQ(1)
            DO I2=1,NQB(2)
              YY=DXL(2)+(I2-0.5D0)*BQ(2)
              YAV=DQ(I1,I2,I3)
              YAV2=DQ2(I1,I2,I3)
              YERR=3.0D0*SQRT(ABS(YAV2-YAV**2*DF))
              YAV=MAX(YAV*DF/VOXEL,1.0D-35)
              YERR=MAX(YERR*DF/VOXEL,1.0D-35)
              WRITE(9,9426) XX,YY,ZZ,YAV,YERR,I1,I2,I3
            ENDDO
            WRITE(9,*) '   '
          ENDDO
          WRITE(9,*) '   '
        ENDDO
 9426   FORMAT(1P,3E11.3,1P,E13.5,E9.1,3I4)
        CLOSE(9)
C
C  ****  Dose distributions at the central axes.
C
        I1C=(NQB(1)/2)+1
        I2C=(NQB(2)/2)+1
        I3C=(NQB(3)/2)+1
C
        IF(NQB(1).GT.1) THEN
          OPEN(9,FILE='psf-x-map.dat')
          WRITE(9,9440)
 9440     FORMAT(
     1       1X,'#  Results from PENMAIN.',
     1      /1X,'#  Q distribution along the central X axis',
     1      /1X,'#  1st column: x (cm).',
     1      /1X,'#  2nd column: Q (1/cm^3).',
     1      /1X,'#  3rd column: statistical uncertainty (3 sigma).')
          DO I1=1,NQB(1)
            XYZ=DXL(1)+(I1-0.5D0)*BQ(1)
            YAV=DQ(I1,I2C,I3C)
            YAV2=DQ2(I1,I2C,I3C)
            YERR=3.0D0*SQRT(ABS(YAV2-YAV**2*DF))
            YAV=YAV*DF/VOXEL
            YERR=YERR*DF/VOXEL
            WRITE(9,'(1X,1P,3E14.6)')
     1        XYZ,MAX(YAV,1.0D-35),MAX(YERR,1.0D-35)
          ENDDO
          CLOSE(9)
        ENDIF
C
        IF(NQB(2).GT.1) THEN
          OPEN(9,FILE='psf-y-map.dat')
          WRITE(9,9450)
 9450     FORMAT(
     1       1X,'#  Results from PENMAIN.',
     1      /1X,'#  Q distribution along the central Y axis',
     1      /1X,'#  1st column: y (cm).',
     1      /1X,'#  2nd column: Q (1/cm^3).',
     1      /1X,'#  3rd column: statistical uncertainty (3 sigma).')
          DO I2=1,NQB(2)
            XYZ=DXL(2)+(I2-0.5D0)*BQ(2)
            YAV=DQ(I1C,I2,I3C)
            YAV2=DQ2(I1C,I2,I3C)
            YERR=3.0D0*SQRT(ABS(YAV2-YAV**2*DF))
            YAV=YAV*DF/VOXEL
            YERR=YERR*DF/VOXEL
            WRITE(9,'(1X,1P,3E14.6)')
     1        XYZ,MAX(YAV,1.0D-35),MAX(YERR,1.0D-35)
          ENDDO
          CLOSE(9)
        ENDIF
C
        IF(NQB(3).GT.1) THEN
          OPEN(9,FILE='psf-z-map.dat')
          WRITE(9,9460)
 9460     FORMAT(
     1       1X,'#  Results from PENMAIN.',
     1      /1X,'#  Q distribution along the central Z axis',
     1      /1X,'#  1st column: z (cm).',
     1      /1X,'#  2nd column: Q (1/cm^3).',
     1      /1X,'#  3rd column: statistical uncertainty (3 sigma).')
          DO I3=1,NQB(3)
            XYZ=DXL(3)+(I3-0.5D0)*BQ(3)
            YAV=DQ(I1C,I2C,I3)
            YAV2=DQ2(I1C,I2C,I3)
            YERR=3.0D0*SQRT(ABS(YAV2-YAV**2*DF))
            YAV=YAV*DF/VOXEL
            YERR=YERR*DF/VOXEL
            WRITE(9,'(1X,1P,3E14.6)')
     1        XYZ,MAX(YAV,1.0D-35),MAX(YERR,1.0D-35)
          ENDDO
          CLOSE(9)
        ENDIF
C
        RETURN
      END
