C
C  Compile PENMAIN with the present 'source.f' file and 'pennucm.f'.
C  Notice that the files 'source.f' and 'pennucm.f' must be copied
C  into the same directory as the PENMAIN/PENELOPE source files.
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                      C
C  PENELOPE/PENNUC (version 2018)                                      C
C  Copyright (c) 2016-2018                                             C
C  Ciemat, Universitat de Barcelona                                    C
C                                                                      C
C  Permission to use, copy, modify, distribute and sell this software  C
C  and its documentation for any purpose is hereby granted without     C
C  fee, provided that the above copyright notice appears in all        C
C  copies and that both that copyright notice and this permission      C
C  notice appear in all supporting documentation. The Ciemat and the   C
C  Universitat de Barcelona make no representations about the suita-   C
C  bility of this software for any purpose. It is provided "as is"     C
C  without express or implied warranty.                                C
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C

      INCLUDE 'pennucm.f'

C  *********************************************************************
C                       SUBROUTINE SOURCE
C  *********************************************************************
      SUBROUTINE SOURCE(METAST)
C
C  >>>>>> Simulation of radioactive sources (with a single nuclide)
C         using the subroutine package PENNUC.F.
C
C  The program PENMAIN allows sources of primary radiations defined by
C  the user. When KPARP (type of primary particle) is set to 0, the
C  initial states of primary particles are generated by the present
C  subroutine, which simulates a radioactive source containing a single
C  nuclide.
C
C  The simulation of nuclear decays is initialised by calling SOURCE0,
C  which is defined as an ENTRY (see below). The PENNUC subroutines
C  read the decay data file and initialise the random sampling. They
C  also provide the value of EPMAX (the highest energy of emitted
C  particles), which is used by the PENELOPE initialisation subroutine.
C
C  After initialisation, each call to SOURCE sets the initial states of
C  the particles released in a decay cascade (a single emission event).
C  One of the particles is stored directly in module TRACK, the
C  others (if any) are sent to the secondary stack. Notice that the
C  particles sent to the stack must be marked with ILB(1)=1, to be
C  treated as proper primary particles by subroutine SHOWER (their
C  initial energies will not be subtracted from the locally deposited
C  energy).
C
C  Input:
C    METAST = 0, initiate the simulation of a new decay,
C           = 1, continue the simulation of a decay that was halted
C                at a metastable level.
C
C  Output flag:
C    METAST = 0, the de-excitation cascade has been completed.
C           = 1, a metastable level has been reached.
C
      USE PENMAIN_mod
      USE PENERROR_mod
      USE TRACK_mod
      USE PENNUC_mod
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      CHARACTER NUCFNAME*45
C
      DIMENSION ILBS(5)
      EXTERNAL RAND
C  ****  Save the position of the decaying nucleus.
      SAVE X0,Y0,Z0
C
      IF(METAST.NE.0) GO TO 100  ! Nucleus in a metastable level.
C
C  ----  Position of the decaying nucleus (same policy as in PENMAIN).
C
      IF(LEXSRC) THEN  ! Extended source.
        IF(LEXBD) THEN  ! Active bodies have been defined.
          NTRIAL=0
 1        CONTINUE
          X=SX0+(RAND(1.0D0)-0.5D0)*SSX
          Y=SY0+(RAND(2.0D0)-0.5D0)*SSY
          Z=SZ0+(RAND(3.0D0)-0.5D0)*SSZ
          CALL LOCATE
          NTRIAL=NTRIAL+1
          IF(NTRIAL.GT.100) THEN
            WRITE(26,'(3X,''WARNING: the sampling of initial '',
     1        ''positions may be very inefficient.'')')
            WRITE(6,'(3X,''WARNING: the sampling of initial '',
     1        ''positions may be very inefficient.'')')
          ENDIF
          IF(IXSBOD(IBODY).EQ.0) GO TO 1
          X0=X
          Y0=Y
          Z0=Z
        ELSE
          X0=SX0+(RAND(1.0D0)-0.5D0)*SSX
          Y0=SY0+(RAND(2.0D0)-0.5D0)*SSY
          Z0=SZ0+(RAND(3.0D0)-0.5D0)*SSZ
        ENDIF
      ELSE  ! Point source.
        X0=SX0
        Y0=SY0
        Z0=SZ0
      ENDIF
C
C  ************  Set the initial states of particles released in a new
C                cascade of the nuclear decay.
C
 100  CONTINUE
C  ----  Common variables (position, weight, ...).
      X=X0
      Y=Y0
      Z=Z0
      IBODY=0  ! Not relevant.
      MAT=0  ! Not relevant.
      PAGE=0.0D0
      WGHT=1.0D0
      IPOL=0  ! Unpolarised photons.
      ISOT=1  ! A single isotope by default.
      CALL PENNUC(ISOT,METAST,IER)
      IF(IER.NE.0) THEN
        WRITE(6,2001) ISOT
 2001 FORMAT(1X,'The intended isotope, ISOT =',I3,' is not loaded.')
        REASON='SOURCE: PENNUC sampling failed.'
        IRETRN=3202
        RETURN
      ENDIF
C
      NSAVED=0
      AGENR1=0.0D0
      IF(NR.GT.0) THEN
        DO I=1,NR
          IF(KPNR(I).GT.0.AND.KPNR(I).LT.4) THEN
            IF(ENR(I).GT.50.0D0) THEN
              NSAVED=NSAVED+1
              IF(NSAVED.EQ.1) THEN
                E=ENR(I)
                U=UNR(I)
                V=VNR(I)
                W=WNR(I)
                KPAR=KPNR(I)
                ILB(1)=1
                ILB(2)=0
                ILB(3)=0
                ILB(4)=ITNR(I)
                ILB(5)=0
                AGENR1=AGENR(I)
              ELSE
                ES=ENR(I)
                US=UNR(I)
                VS=VNR(I)
                WS=WNR(I)
                KPARS=KPNR(I)
                ILBS(1)=1
                ILBS(2)=0
                ILBS(3)=0
                ILBS(4)=ITNR(I)
                ILBS(5)=0
                IF(LAGE) PAGE=AGENR(I)  ! Transferred through TRACK_mod.
                CALL STORES(ES,X,Y,Z,US,VS,WS,WGHT,KPARS,ILBS,IPOL)
                IF(IRETRN.NE.0) RETURN
              ENDIF
            ENDIF
          ENDIF
        ENDDO
      ENDIF
      IF(LAGE) PAGE=AGENR1  ! Transferred through TRACK_mod.
C
      IF(NSAVED.EQ.0) THEN
        E=0.0D0  ! Tells the main program to generate a new decay.
        U=1.0D0
        V=0.0D0
        W=0.0D0
        KPAR=1
        ILB(1)=1
        ILB(2)=0
        ILB(3)=0
        ILB(4)=0
        ILB(5)=0
        PAGE=0.0D0
      ENDIF
      RETURN
C
C  ************  PENNUC initialisation.
C
      ENTRY SOURCE0(NUCFNAME)
C
C  ****  Read nuclear decay data from the NUCLEIDE file and
C        initialise the random sampling.
C
      IWR=15
      OPEN(IWR,FILE='pennuc.dat')
      CALL PENNUC0(NUCFNAME,EPMAX,IWR,IER)
      CLOSE(IWR)
      EPMAX=1.001D0*EPMAX
C  ----  The parameter EPMAX is transferred to the main program thorugh
C        the PENMAIN_mod module.
      IF(IER.NE.0) THEN
        WRITE(6,*) 'SOURCE0: PENNUC initialisation failed.'
        REASON='SOURCE0: PENNUC initialisation failed.'
        IRETRN=3201
      ENDIF
      RETURN
      END
