CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                      C
C  PENELOPE/PENGEOM (version 2018)                                     C
C  Copyright (c) 2001-2018                                             C
C  Universitat de Barcelona                                            C
C                                                                      C
C  Permission to use, copy, modify, distribute and sell this software  C
C  and its documentation for any purpose is hereby granted without     C
C  fee, provided that the above copyright notice appears in all        C
C  copies and that both that copyright notice and this permission      C
C  notice appear in all supporting documentation. The Universitat de   C
C  Barcelona makes no representations about the suitability of this    C
C  software for any purpose. It is provided 'as is' without express    C
C  or implied warranty.                                                C
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C  *********************************************************************
C                       SUBROUTINE PACKAGE PENFIELD
C  *********************************************************************
C
C                                       Francesc Salvat. December, 2018.
C
C  This subroutine package generates electron/positron trajectories in
C  an external static electromagnetic field. The material structure is
C  described by means of the geometry package PENGEOM.
C
C  The user must provide the following subroutines (see the examples at
C  the end of this file);
C     SUBROUTINE GETEMF(X,Y,Z,EX,EY,EZ,BX,BY,BZ)
C     SUBROUTINE GETSCP(X,Y,Z,PHI)
C  which deliver the electric field (EX,EY,EZ) in V/cm, the magnetic
C  field (BX,BY,BZ) in G, and the scalar potential PHI in V, at an
C  arbitrary point with coordinates (X,Y,Z) cm.
C
C  Notice that 1 G = 1.0E-4 T and 1 statV = 299.792458 V.
C
C  The sequence of calls to generate a particle track is the following,
C  1: CALL START  ! Initialises the simulation of the track.
C  2: CALL TPEMF0(ULDV,ULDE,ULEM,DSMAX)  ! Determines the maximum
C       ! step length DSMAX consistent with the adopted delta values.
C  3: CALL JUMP(DSMAX,DS)  ! Determines the path length DS to the next
C       ! interaction event.
C  4: CALL TPEMF1(DS,DSEF,NCROSS)  ! Moves the particle to the
C       ! end of the step and determines its final direction and energy.
C  5: CALL KNOCK(DE,ICOL)  ! Simulates next interaction. Interaction
C       !  energy losses are assumed to be deposited at the hinge.
C  6: GO TO 1 or 2.
C  This sequence has to be discontinued when the particle leaves the
C  material system or is absorbed.
C
C  Although the transport of photons is not affected by the field, the
C  present subroutines can also be called to generate photon histories.
C  This is useful to simplify the structure of the main program (all
C  kinds of particles can be tracked by using the same sequence of calls
C  to the simulation subroutines).
C
C  IMPORTANT NOTE: One of the practical features of PENGEOM is that
C  particles are automatically transported through void regions (i.e.
C  those with MAT=0) following straight trajectories without loosing
C  energy. Of course, in the presence of fields this would give
C  erroneous results. The easiest method to avoid this conflict is to
C  fill up the voids with a fake material of very small density.
C
C  *********************************************************************
C                       SUBROUTINE TPEMF1
C  *********************************************************************
      SUBROUTINE TPEMF1(DS,DSEF,NCROSS)
C
C  This subroutine generates particle trajectory segments in static
C  electromagnetic fields. It is assumed that subroutine TPEMF0 has
C  been called immediately before entering TPEMF1.
C
C  Input/output (through module TRACK_mod):
C     X, Y, Z ... position coordinates,
C     U, V, W ... direction cosines of the direction vector,
C     IBODY ..... body where the point (X,Y,Z) is located,
C     MAT ....... material in body IBODY.
C
C  Input values (arguments):
C     DS ........ path length to travel,
C
C  Output values (arguments):
C     DSEF....... travelled curved path length before leaving the
C                 initial material or completing the jump (less
C                 than DS if the track crosses an interface),
C     NCROSS .... = 0 if the whole step is contained in the
C                     initial material,
C                 .gt.0 if the particle has crossed an interface,
C                       i.e. if it has entered a new material.
C
      USE TRACK_mod
      USE PENELOPE_mod
      USE PENERROR_mod
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      PARAMETER (REV=5.10998928D5)  ! Electron rest energy (eV)
      PARAMETER (SLCM=2.99792458D10)  ! Speed of light (cm/s)
      PARAMETER (SLCM2=SLCM*SLCM)
      PARAMETER (C300=299.792458D0)  ! Conversion factor.
C  ****  Energy grid and interpolation constants for the current energy.
      COMMON/CEGRID/EMIN,EL,EU,ET(NEGP),DLEMP(NEGP),DLEMP1,DLFC,
     1  XEL,XE,XEK,KE
      COMMON/COMFLD/CHARGE,ETOT,BETA2,CONSE,BETA,SPEED,
     1              EX,EY,EZ,BX,BY,BZ,BF2,EDOTV
C
      IF(KPAR.EQ.2) THEN
        CALL STEP(DS,DSEF,NCROSS)
        RETURN
      ENDIF
C
C  ****  Displacement due to inertia and electric force.
C
      HDS2=0.5D0*DS*DS
      DXX=CONSE*(EX-BETA2*EDOTV*U)
      DRX=DS*U+HDS2*DXX
      DVX=DS*DXX  ! Velocity variation components in units of SPEED.
      DYY=CONSE*(EY-BETA2*EDOTV*V)
      DRY=DS*V+HDS2*DYY
      DVY=DS*DYY
      DZZ=CONSE*(EZ-BETA2*EDOTV*W)
      DRZ=DS*W+HDS2*DZZ
      DVZ=DS*DZZ
C
C  ****  Displacement due to magnetic field.
C
      IF(BF2.GT.1.0D-32) THEN
        B=SQRT(BF2)
        RB=1.0D0/B
        OMEGA=ABS(CHARGE)*SLCM*C300*B/ETOT
        IF(CHARGE.GT.0.0D0) THEN
          OX=-BX*RB  ! O=OMEGA/|OMEGA|
          OY=-BY*RB
          OZ=-BZ*RB
        ELSE
          OX=BX*RB
          OY=BY*RB
          OZ=BZ*RB
        ENDIF
        VDOTO=U*OX+V*OY+W*OZ
        V0PX=U-VDOTO*OX  ! V_perp/V.
        V0PY=V-VDOTO*OY
        V0PZ=W-VDOTO*OZ
        V0OX=V0PY*OZ-V0PZ*OY  ! (V_perp/V) vecprod O.
        V0OY=V0PZ*OX-V0PX*OZ
        V0OZ=V0PX*OY-V0PY*OX
C
        ARG=DS*OMEGA/SPEED
        F1=COS(ARG)-1.0D0
        F2=SIN(ARG)
        FACT=SPEED/OMEGA
        DRX=DRX-DS*V0PX+(F1*V0OX+F2*V0PX)*FACT
        DRY=DRY-DS*V0PY+(F1*V0OY+F2*V0PY)*FACT
        DRZ=DRZ-DS*V0PZ+(F1*V0OZ+F2*V0PZ)*FACT
        DVX=DVX+F1*V0PX-F2*V0OX  ! in units of SPEED.
        DVY=DVY+F1*V0PY-F2*V0OY
        DVZ=DVZ+F1*V0PZ-F2*V0OZ
      ENDIF
C
C  ****  Combined displacement through the geometrical system.
C
      X0=X
      Y0=Y
      Z0=Z
      U0=U
      V0=V
      W0=W
      DSR=SQRT(DRX**2+DRY**2+DRZ**2)
      IF(DSR.GT.1.0D-30) THEN
        U=DRX/DSR
        V=DRY/DSR
        W=DRZ/DSR
        CALL STEP(DSR,DSEF,NCROSS)
      ELSE
        WRITE(6,1000)
 1000 FORMAT(1X,'*** Possible inconsistency in subroutine TPEMF1.')
        REASON='Possible inconsistency in subroutine TPEMF1.'
        IRETRN=1501
        RETURN
      ENDIF
C
C  ****  Direction cosines at the end of the substep.
C
      FACT=DSEF/DSR
      U=U0+FACT*DVX
      V=V0+FACT*DVY
      W=W0+FACT*DVZ
      DSEF=FACT*DS
      FNORM=U*U+V*V+W*W
      IF(ABS(FNORM-1.0D0).GT.1.0D-13) THEN
        FNORM=1.0D0/SQRT(FNORM)
        U=U*FNORM
        V=V*FNORM
        W=W*FNORM
      ENDIF
C
C  ****  Energy at the end of the substep.
C
      CALL GETSCP(X0,Y0,Z0,PHI0)
      CALL GETSCP(X,Y,Z,PHI)
      E0=E
      E=E+CHARGE*(PHI0-PHI)
C  ****  Kinetic energies less than EABS cannot be tracked.
      IF(E.LT.EABS(KPAR,MAT)) E=0.0D0
C
      IF(ABS(E-E0).GT.0.1D0) THEN
        XEL=LOG(E)
        XE=1.0D0+(XEL-DLEMP1)*DLFC
        KE=XE
        XEK=XE-KE
      ENDIF
C
      RETURN
      END
C  *********************************************************************
C                       SUBROUTINE TPEMF0
C  *********************************************************************
      SUBROUTINE TPEMF0(ULDV,ULDE,ULEM,DSM)
C
C  This subroutine determines the maximum allowed step length of elec-
C  trons/positrons in the electromagnetic field, consistent with the
C  adopted delta values.
C
C  Input values (arguments):
C    ULDV ... upper limit on the amount of deflection over the step
C             due to the em-field (0.02, typical).
C    ULDE ... upper limit on the relative energy variation over the
C             step in the em-field (0.02, typical).
C    ULEM ... upper limit on the amount of change of the em-field
C             over the step (0.02, typical).
C
C  Output values (arguments):
C    DSM ... maximum step length consistent with accuracy limits.
C
      USE TRACK_mod
      USE PENELOPE_mod
      USE PENERROR_mod
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      PARAMETER (REV=5.10998928D5)  ! Electron rest energy (eV)
      PARAMETER (TREV=REV+REV)
      PARAMETER (SLCM=2.99792458D10)  ! Speed of light (cm/s)
      PARAMETER (C300=299.792458D0)  ! Conversion factor.
      COMMON/COMFLD/CHARGE,ETOT,BETA2,CONSE,BETA,SPEED,
     1              EX,EY,EZ,BX,BY,BZ,BF2,EDOTV
C
      IF(KPAR.EQ.2) RETURN
      DSM=1.0D35
      CHARGE=DBLE(KPAR-2)
C  ****  Ensure normalisation of the direction vector.
      DXYZ=U*U+V*V+W*W
      IF(ABS(DXYZ-1.0D0).GT.1.0D-14) THEN
        FNORM=1.0D0/SQRT(DXYZ)
        U=FNORM*U
        V=FNORM*V
        W=FNORM*W
      ENDIF
C
      ETOT=E+REV
      IF(E.GT.1.0D3) THEN
        BETA2=E*(E+TREV)/ETOT**2
        CONSE=CHARGE/(ETOT*BETA2)
      ELSE
        TECOR=2.0D0*MAX(E,0.01D0)
        BETA2=TECOR/REV
        CONSE=CHARGE/TECOR
      ENDIF
      BETA=SQRT(BETA2)
      SPEED=BETA*SLCM
C
      CALL GETEMF(X,Y,Z,EX,EY,EZ,BX,BY,BZ)
      EF2=EX**2+EY**2+EZ**2
      BF2=BX**2+BY**2+BZ**2
C
C  ****  Step-size truncation based upon energy change
C        by the external field.
C
      EDOTV=EX*U+EY*V+EZ*W
      AEDOTV=ABS(EDOTV)
      IF(AEDOTV.GT.1.0D-30) DSM=MIN(DSM,MAX(ULDE*E,0.01D0)/AEDOTV)
C  ****  If the particle is at rest (approx.) it starts moving in the
C        direction of the electric field.
      IF(E.LT.0.101D0) THEN
        IF(EF2.GT.0.9D0*EABS(KPAR,MAT)) THEN
          EDOTV=SQRT(CHARGE**2*EF2)
          REF=CHARGE/EDOTV
          U=REF*EX
          V=REF*EY
          W=REF*EZ
        ELSE
          REASON='The electron came to rest with no E-field.'
          IRETRN=1502
          RETURN
        ENDIF
      ENDIF
C
C  ****  Step-size truncation based upon direction vector change
C        by the external field.
C
      CTB=BETA*C300
      DVX=EX-BETA2*EDOTV*U+CTB*(V*BZ-W*BY)
      DVY=EY-BETA2*EDOTV*V+CTB*(W*BX-U*BZ)
      DVZ=EZ-BETA2*EDOTV*W+CTB*(U*BY-V*BX)
      DV=SQRT(DVX*DVX+DVY*DVY+DVZ*DVZ)*ABS(CONSE)
      IF(DV.GT.1.0D-30) DSM=MIN(DSM,ULDV/DV)
C
C  ****  Step-size truncation based upon keeping the external field
C        approximately a constant.
C
      IF(DSM.GT.1.0D-8) THEN
        XT=X+DSM*U
        YT=Y+DSM*V
        ZT=Z+DSM*W
        CALL GETEMF(XT,YT,ZT,EXT,EYT,EZT,BXT,BYT,BZT)
C
        VAREF2=(EXT-EX)**2+(EYT-EY)**2+(EZT-EZ)**2
        IF(VAREF2.GT.1.0D-6*EF2) THEN
          IF(EF2.GT.0.01D0) DSM=MIN(DSM,ULEM*DSM*DSQRT(EF2/VAREF2))
        ENDIF
C
        VARBF2=(BXT-BX)**2+(BYT-BY)**2+(BZT-BZ)**2
        IF(VARBF2.GT.1.0D-6*BF2) THEN
          IF(BF2.GT.0.01D0) DSM=MIN(DSM,ULEM*DSM*DSQRT(BF2/VARBF2))
        ENDIF
      ELSE
        DSM=1.0D-8
      ENDIF
      RETURN
      END
C  *********************************************************************
C                       SUBROUTINE GETEMF
C  *********************************************************************
      SUBROUTINE GETEMF(XP,YP,ZP,EX,EY,EZ,BX,BY,BZ)
C
C  This subroutine returns the electric and magnetic field components
C  at the input point (XP,YP,ZP).
C
C  Input arguments:
C  (XP,YP,ZP) ...  position coordinates (cm).
C
C  Output arguments:
C  (EX,EY,EZ) ...  components of the E-field, in V/cm.
C  (BX,BY,BZ) ...  components of the B-field, in G.
C
C  Notice that 1 G = 1.0E-4 T and 1 statV = 299.792458 V.
C
C  It is assumed that users will provide an equivalent subroutine, with
C  the same name and arguments, to define their fields.
C
C  For example:
C  For a uniform E-field: (EX,EY,EZ) three constants.
C                         (BX,BY,BZ)=(0,0,0).
C  For a uniform B-field: (EX,EY,EZ)=(0,0,0).
C                         (BX,BY,BZ) three constants.
C  For a point charge at rest at the origin of coordinates:
C                         EX=C*XP/(SQRT(XP**2+YP**2+ZP**2))**3,
C                         EY=C*YP/(SQRT(XP**2+YP**2+ZP**2))**3,
C                         EZ=C*ZP/(SQRT(XP**2+YP**2+ZP**2))**3,
C                         (BX,BY,BZ)=(0,0,0).
C                         where C is a constant.
C
C  This example is for a uniform EM field, the components of the E-field
C  and of the B-field are entered through the common block /UFIELD/.
C
      USE TRACK_mod
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      COMMON/UFIELD/EFX,EFY,EFZ,BFX,BFY,BFZ
C
C  ****  The field is set equal to zero in regions where MAT=0, i.e. in
C        the outer vacuum, because otherwise particles that leave the
C        system could return to it under the action of the field.
C
      IF(MAT.EQ.0) THEN
        EX=0.0D0
        EY=0.0D0
        EZ=0.0D0
        BX=0.0D0
        BY=0.0D0
        BZ=0.0D0
        RETURN
      ENDIF
C
      EX=EFX
      EY=EFY
      EZ=EFZ
      BX=BFX
      BY=BFY
      BZ=BFZ
      RETURN
      END
C  *********************************************************************
C                       SUBROUTINE GETSCP
C  *********************************************************************
      SUBROUTINE GETSCP(XP,YP,ZP,PHI)
C
C  This routine returns the scalar potential at the point (XP,YP,ZP).
C
C  Input arguments:
C  (XP,YP,ZP) ...  position coordinates (cm).
C
C  Output argument:
C  PHI ..........  scalar potential, in V.
C
C  Notice that 1 statV = 299.792458 V.
C
C  It is assumed that users will provide an equivalent subroutine, with
C  the same name and arguments, to define their fields.
C
C  For example:
C  For a uniform E-field: (EX,EY,EZ) three constants.
C                         PHI=-(EX*XP+EY*YP+EZ*ZP).
C  For a point charge at rest at the origin of coordinates:
C                         PHI=C/SQRT(XP**2+YP**2+ZP**2),
C                         where C is a constant.
C
C  This example is for a uniform EM field. The components of the E-field
C  are entered through the common block /UFIELD/.
C
      USE TRACK_mod
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      COMMON/UFIELD/EFX,EFY,EFZ,BFX,BFY,BFZ
C
C  ****  The potential is set equal to zero in regions where MAT=0,
C        i.e. in the outer vacuum, because otherwise particles that
C        leave the system could return to it under the action of the
C        field.
C
      IF(MAT.EQ.0) THEN
        PHI=0.0D0
        RETURN
      ENDIF
C
      PHI=-(EFX*XP+EFY*YP+EFZ*ZP)
      RETURN
      END
